"rubrik: A Python package for interacting with the Rubrik CDM API."

from .rubrik_cdm import Connect
from .rubrik_cdm import Bootstrap

import logging

# Define the logging params
console_output_handler = logging.StreamHandler()
formatter = logging.Formatter("[%(asctime)s] [%(levelname)s] -- %(message)s")
console_output_handler.setFormatter(formatter)

log = logging.getLogger(__name__)
log.addHandler(console_output_handler)

__version__ = "1.0"
__author__ = "Rubrik Ranger Team"
__all__ = []



# Copyright 2018 Rubrik, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License prop
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Rubrik SDK Cluster class.
"""

import sys
from .api import Api

_API = Api


class Cluster(_API):
    """This class contains methods related to the managment of the Rubrik cluster itself.
    """

    def cluster_version(self, timeout=15):
        """Retrieves the software version of the Rubrik cluster.

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})

        Returns:
            str -- The version of CDM installed on the Rubrik cluster.
        """

        self.log('cluster_version: Getting the software version of the Rubrik cluster.')
        return self.get('v1', '/cluster/me/version', timeout)['version']

    def cluster_node_ip(self, timeout=15):
        """Retrive the IP Address for each node in the Rubrik cluster.

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})

        Returns:
            list -- A list that contains the IP Address for each node in the Rubrik cluster.
        """

        self.log('cluster_node_ip: Generating a list of all Cluster Node IPs.')
        api_request = self.get('internal', '/cluster/me/node', timeout)

        node_ip_list = []

        for node in api_request['data']:
            node_ip_list.append(node["ipAddress"])

        return node_ip_list

    def end_user_authorization(self, object_name, end_user, object_type='vmware', timeout=15):
        """Grant an End User authorization to the provided object.

        Arguments:
            object_name {str} -- The name of the object you wish to grant the `end_user` authorization to.
            end_user {str} -- The name of the end user you wish to grant authorization to.

        Keyword Arguments:
            object_type {str} -- The Rubrik object type you wish to backup. (default: {vmware}) (choices: {vmware}) 
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})

        Returns:
            str -- No change required. The End User "`end_user`" is already authorized to interact with the "`object_name`" VM.
            dict -- The API response from `POST /internal/authorization/role/end_user`.
        """

        valid_object_type = ['vmware']

        if object_type not in valid_object_type:
            sys.exit("Error: The end_user_authorization() object_type argument must be one of the following: {}.".format(
                valid_object_type))

        self.log("end_user_authorization: Searching the Rubrik cluster for the vSphere VM '{}'.".format(object_name))
        vm_id = self.object_id(object_name, object_type)

        self.log("end_user_authorization: Searching the Rubrik cluster for the End User '{}'.".format(end_user))
        user = self.get('internal', '/user?username={}'.format(end_user))

        if not user:
            sys.exit('The Rubrik cluster does not contain a End User account named "{}".'.format(end_user))
        else:
            user_id = user[0]['id']

        self.log("end_user_authorization: Searching the Rubrik cluster for the End User '{}' authorizations.".format(end_user))
        user_authorization = self.get('internal', '/authorization/role/end_user?principals={}'.format(user_id))

        authorized_objects = user_authorization['data'][0]['privileges']['restore']

        if vm_id in authorized_objects:
            return 'No change required. The End User "{}" is already authorized to interact with the "{}" VM.'.format(end_user, object_name)
        else:
            config = {}
            config['principals'] = [user_id]
            config['privileges'] = {"restore": [vm_id]}

            return self.post('internal', '/authorization/role/end_user', config, timeout=timeout)



# Copyright 2018 Rubrik, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License prop
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Rubrik SDK Data_Management class.
"""

import sys
import re
from .api import Api

_API = Api


class Data_Management(_API):
    """This class contains methods related to backup and restore operations for the various objects managed by the Rubrik cluster."""

    def on_demand_snapshot(self, object_name, object_type, sla_name='current'):
        """Initiate an on-demand snapshot.

        Arguments:
            object_name {str} -- The name of the Rubrik object to take a on-demand snapshot of.
            object_type {str} -- The Rubrik object type you want to backup. (choices: {vmware})

        Keyword Arguments:
            sla_name {str} -- The SLA Domain name you want to assign the on-demand snapshot to. By default, the currently assigned SLA Domain will be used. (default: {'current'})

        Returns:
            tuple -- The full API response for `POST /v1/vmware/vm/{ID}/snapshot` and the job status URL which can be used to monitor progress of the snapshot. (api_response, job_status_url)
        """

        valid_object_type = ['vmware']

        if object_type not in valid_object_type:
            sys.exit("Error: The on_demand_snapshot() object_type argument must be one of the following: {}.".format(valid_object_type))

        if object_type is 'vmware':
            self.log("on_demand_snapshot: Searching the Rubrik cluster for the vSphere VM '{}'.".format(object_name))
            vm_id = self.object_id(object_name, object_type)

            if sla_name is 'current':
                self.log(
                    "on_demand_snapshot: Searching the Rubrik cluster for the SLA Domain assigned to the vSphere VM '{}'.".format(object_name))

                vm_summary = self.get('v1', '/vmware/vm/{}'.format(vm_id))
                sla_id = vm_summary['effectiveSlaDomainId']

            elif sla_name is not 'current':
                self.log("on_demand_snapshot: Searching the Rubrik cluster for the SLA Domain '{}'.".format(sla_name))
                sla_id = self.object_id(sla_name, 'sla')

            config = {}
            config['slaId'] = sla_id

            self.log("on_demand_snapshot: Initiating snapshot for the vSphere VM '{}'.".format(object_name))
            api_request = self.post('v1', '/vmware/vm/{}/snapshot'.format(vm_id), config)

            snapshot_status_url = api_request['links'][0]['href']

        return (api_request, snapshot_status_url)

    def object_id(self, object_name, object_type):
        """Get the ID of a Rubrik object by providing its name.

        Arguments:
            object_name {str} -- The name of the Rubrik object whose ID you wish to lookup.
            object_type {str} -- The object type you wish to look up. (choices: {vmware, sla, vmware_host})

        Returns:
            str -- The ID of the provided Rubrik object.
        """

        valid_object_type = ['vmware', 'sla', 'vmware_host']

        if object_type not in valid_object_type:
            sys.exit("Error: The object_id() object_type argument must be one of the following: {}.".format(valid_object_type))

        if object_type is 'vmware':
            object_summary_api_version = 'v1'
            object_summary_api_endpoint = '/vmware/vm?primary_cluster_id=local&is_relic=false&name={}'.format(
                object_name)
        elif object_type is 'sla':
            object_summary_api_version = 'v1'
            object_summary_api_endpoint = '/sla_domain?primary_cluster_id=local&name={}'.format(object_name)
        elif object_type is 'vmware_host':
            object_summary_api_version = 'v1'
            object_summary_api_endpoint = '/vmware/host?primary_cluster_id=local'

        self.log("object_id: Getting the object id for the {} object '{}'.".format(object_type, object_name))
        api_request = self.get(object_summary_api_version, object_summary_api_endpoint)

        if api_request['total'] == 0:
            sys.exit("Error: The {} object '{}' was not found on the Rubrik cluster.".format(object_type, object_name))
        elif api_request['total'] > 0:
            for item in api_request['data']:
                if item['name'] == object_name:
                    object_id = item['id']
                    return object_id

            sys.exit("Error: The {} object '{}' was not found on the Rubrik cluster.".format(object_type, object_name))

    def assign_sla(self, object_name, sla_name, object_type, timeout=30):
        """Assign a Rubrik object to an SLA Domain.

        Arguments:
            object_name {str} -- The name of the Rubrik object you wish to assign to an SLA Domain. To exclude the object from all SLA assignments use `do not protect` as the `sla_name`. To assign the selected object to the SLA of the next higher level object use `clear` as the `sla_name`.
            sla_name {str} -- The name of the SLA Domain you wish to assign an object to.
            object_type {str} -- The Rubrik object type you want to assign to the SLA Domain. (choices: {vmware})

        Keyword Arguments:
            timeout {str} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {30})

        Returns:
            str -- No change required. The vSphere VM '`object_name`' is already assigned to the '`sla_name`' SLA Domain.
            dict -- The full API reponse for `POST /internal/sla_domain/{sla_id}/assign`.
        """

        valid_object_type = ['vmware']

        if object_type not in valid_object_type:
            sys.exit("Error: The assign_sla() object_type argument must be one of the following: {}.".format(valid_object_type))

        # Determine if 'do not protect' or 'clear' are the SLA Domain Name
        do_not_protect_regex = re.findall('\\bdo not protect\\b', sla_name, flags=re.IGNORECASE)
        clear_regex = re.findall('\\bclear\\b', sla_name, flags=re.IGNORECASE)

        if len(do_not_protect_regex) > 0:
            sla_id = "UNPROTECTED"
        elif len(clear_regex) > 0:
            sla_id = 'INHERIT'
        else:
            self.log("assign_sla: Searching the Rubrik cluster for the SLA Domain '{}'.".format(sla_name))
            sla_id = self.object_id(sla_name, 'sla')

        if object_type is 'vmware':
            self.log("assign_sla: Searching the Rubrik cluster for the vSphere VM '{}'.".format(object_name))
            vm_id = self.object_id(object_name, object_type)

            self.log("assign_sla: Determing the SLA Domain currently assigned to the vSphere VM '{}'.".format(object_name))
            vm_summary = self.get('v1', '/vmware/vm/{}'.format(vm_id))
            if sla_id is "INHERIT":
                current_sla_id = vm_summary['configuredSlaDomainId']
            else:
                current_sla_id = vm_summary['effectiveSlaDomainId']

            if sla_id == current_sla_id:
                return "No change required. The vSphere VM '{}' is already assigned to the '{}' SLA Domain.".format(object_name, sla_name)
            else:
                self.log("assign_sla: Assigning the vSphere VM '{}' to the '{}' SLA Domain.".format(object_name, sla_name))

                config = {}
                config['managedIds'] = [vm_id]

                return self.post('internal', '/sla_domain/{}/assign'.format(sla_id), config, timeout)

    def vsphere_live_mount(self, vm_name, date='latest', time='latest', host='current', remove_network_devices=False, power_on=True):
        """Live Mount a vSphere VM from a specified snapshot. If a specific date and time is not provided, the last snapshot taken will be used.

        Arguments:
            vm_name {str} -- The name of the vSphere VM to Live Mount.

        Keyword Arguments:
            date {str} -- The date of the snapshot you wish to Live Mount formated as `Month-Day-Year` (ex: 1-15-2014). If `latest` is specified, the last snapshot taken will be used. (default: {'latest'})
            time {str} -- The time of the snapshot you wish to Live Mount formated formated as `Hour:Minute AM/PM` (ex: 1:30 AM). If `latest` is specified, the last snapshot taken will be used. (default: {'latest'})
            host {str} -- The hostname or IP address of the ESXi host to Live Mount the VM on. By default, the current host will be used. (default: {'current'})
            remove_network_devices {bool} -- Flag that determines whether to remove the network interfaces from the Live Mounted VM. Set to `True` to remove all network interfaces. (default: {False})
            power_on {bool} -- Flag that determines whether the VM should be powered on after the Live Mount. Set to `True` to power on the VM. Set to `False` to mount the VM but not power it on. (default: {True})

        Returns:
            dict -- The full response of `POST /v1/vmware/vm/snapshot/{snapshot_id}/mount`.
        """

        if isinstance(remove_network_devices, bool) is False:
            sys.exit("Error: The 'remove_network_devices' argument must be True or False.")
        elif isinstance(power_on, bool) is False:
            sys.exit("Error: The 'power_on' argument must be True or False.")
        elif date is not 'latest' and time is 'latest' or date is 'latest' and time is not 'latest':
            sys.exit("Error: The date and time arguments most both be 'latest' or a specific date and time.")

        self.log("vsphere_live_mount: Searching the Rubrik cluster for the vSphere VM '{}'.".format(vm_name))
        vm_id = self.object_id(vm_name, 'vmware')

        self.log("vsphere_live_mount: Getting a list of all Snapshots for vSphere VM '{}'.".format(vm_name))
        vm_summary = self.get('v1', '/vmware/vm/{}'.format(vm_id))

        if date is 'latest' and time is 'latest':
            number_of_snapshots = len(vm_summary['snapshots'])
            snapshot_id = vm_summary['snapshots'][number_of_snapshots - 1]['id']
        else:
            self.log("vsphere_live_mount: Converting the provided date/time into UTC.")
            snapshot_date_time = self._date_time_conversion(date, time)

            current_snapshots = {}

            for snapshot in vm_summary['snapshots']:
                current_snapshots[snapshot['id']] = snapshot['date']

            self.log("vsphere_live_mount: Searching for the provided snapshot.")
            for id, date_time in current_snapshots.items():
                if snapshot_date_time in date_time:
                    snapshot_id = id

        try:
            snapshot_id
        except NameError:
            sys.exit("Error: The vSphere VM '{}' does not have a snapshot taken on {} at {}.".format(vm_name, date, time))
        else:
            if host is 'current':
                host_id = vm_summary['hostId']
            else:
                host_id = self.object_id(host, 'vmware_host')

            config = {}
            config['hostId'] = host_id
            config['removeNetworkDevices'] = remove_network_devices
            config['powerOn'] = power_on

            self.log("vsphere_live_mount: Live Mounting the snapshot taken on {} at {} for vSphere VM '{}'.".format(
                date, time, vm_name))

            return self.post('v1', '/vmware/vm/snapshot/{}/mount'.format(snapshot_id), config)

    def vsphere_instant_recovery(self, vm_name, date='latest', time='latest', host='current', remove_network_devices=False, power_on=True, disable_network=False, keep_mac_addresses=False, preserve_moid=False):
        """Instantly recover a vSphere VM from a provided snapshot. If a specific date and time is not provided, the last snapshot taken will be used.

        Arguments:
            vm_name {str} -- The name of the VM to Instantly Recover.

        Keyword Arguments:
            date {str} -- The date of the snapshot you wish to Instantly Recover formated as `Month-Day-Year` (ex: 1-15-2014). If 'latest' is specified, the last snapshot taken will used. (default: {'latest'})
            time {str} -- The time of the snapshot you wish to Instantly Recover formated formated as `Hour:Minute AM/PM`  (ex: 1:30 AM). If 'latest' is specified, the last snapshot taken will be used. (default: {'latest'})
            host {str} -- The hostname or IP address of the ESXi host to Instantly Recover the VM on. By default, the current host will be used. (default: {'current'})
            remove_network_devices {bool} -- Flag that determines whether to remove the network interfaces from the Instantly Recovered VM. Set to `True` to remove all network interfaces. (default: {False})
            power_on {bool} -- Flag that determines whether the VM should be powered on after Instant Recovery. Set to `True` to power on the VM. Set to `False` to instantly recover the VM but not power it on. (default: {True})
            disable_network {bool} -- Sets the state of the network interfaces when the VM is instantly recovered. Use `False` to enable the network interfaces. Use `True` to disable the network interfaces. Disabling the interfaces can prevent IP conflicts. (default: {False})
            keep_mac_addresses {bool} -- Flag that determines whether the MAC addresses of the network interfaces on the source VM are assigned to the new VM. Set to `True` to assign the original MAC addresses to the new VM. Set to `False` to assign new MAC addresses. When 'remove_network_devices' is set to `True`, this property is ignored. (default: {False})
            preserve_moid {bool} -- Flag that determines whether to preserve the MOID of the source VM in a restore operation. Use `True` to keep the MOID of the source. Use `False` to assign a new moid. (default: {False})

        Returns:
            dict -- The full response of `POST /v1/vmware/vm/snapshot/{snapshot_id}/instant_recover`.
        """

        if isinstance(remove_network_devices, bool) is False:
            sys.exit("Error: The 'remove_network_devices' argument must be True or False.")
        elif isinstance(power_on, bool) is False:
            sys.exit("Error: The 'power_on' argument must be True or False.")
        elif isinstance(disable_network, bool) is False:
            sys.exit("Error: The 'disable_network' argument must be True or False.")
        elif isinstance(keep_mac_addresses, bool) is False:
            sys.exit("Error: The 'keep_mac_addresses' argument must be True or False.")
        elif isinstance(preserve_moid, bool) is False:
            sys.exit("Error: The 'preserve_moid' argument must be True or False.")
        elif date is not 'latest' and time is 'latest' or date is 'latest' and time is not 'latest':
            sys.exit("Error: The date and time arguments most both be 'latest' or a specific date and time.")

        self.log("vsphere_instant_recovery: Searching the Rubrik cluster for the vSphere VM '{}'.".format(vm_name))
        vm_id = self.object_id(vm_name, 'vmware')

        self.log("vsphere_instant_recovery: Getting a list of all Snapshots for vSphere VM '{}'.".format(vm_name))
        vm_summary = self.get('v1', '/vmware/vm/{}'.format(vm_id))

        if date is 'latest' and time is 'latest':
            number_of_snapshots = len(vm_summary['snapshots'])
            snapshot_id = vm_summary['snapshots'][number_of_snapshots - 1]['id']
        else:
            self.log("vsphere_instant_recovery: Converting the provided date/time into UTC.")
            snapshot_date_time = self._date_time_conversion(date, time)

            current_snapshots = {}
            for snapshot in vm_summary['snapshots']:
                current_snapshots[snapshot['id']] = snapshot['date']

            self.log("vsphere_instant_recovery: Searching for the provided snapshot.")
            for id, date_time in current_snapshots.items():
                if snapshot_date_time in date_time:
                    snapshot_id = id

        try:
            snapshot_id
        except NameError:
            sys.exit("Error: The vSphere VM '{}' does not have a snapshot taken on {} at {}.".format(vm_name, date, time))
        else:
            if host is 'current':
                host_id = vm_summary['hostId']
            else:
                host_id = self.object_id(host, 'vmware_host')

            config = {}
            config['hostId'] = host_id
            config['removeNetworkDevices'] = remove_network_devices
            config['powerOn'] = power_on
            config['disableNetwork'] = disable_network
            config['keepMacAddresses'] = keep_mac_addresses
            config['preserveMoid'] = preserve_moid

            self.log("vsphere_instant_recovery: Instantly Recovering the snapshot taken on {} at {} for vSphere VM '{}'.".format(
                date, time, vm_name))

            return self.post('v1', '/vmware/vm/snapshot/{}/instant_recover'.format(snapshot_id), config)

    def _date_time_conversion(self, date, time):
        """All date values returned by the Rubrik API are stored in Coordinated Universal Time (UTC)
        and need to be converted to the timezone configured on the Rubrik cluster in order to match
        the values provided by the end user in various functions. This internal function will handle that
        conversion process.

        Arguments:
            date {str} -- A date value formated as `Month-Day-Year` (ex: 1/15/2014).
            time {str} -- A time value formated as `Hour:Minute AM/PM` (ex: 1:30 AM).

        Returns:
            str -- A combined date/time value formated as `Year-Month-DayTHour:Minute` where Hour:Minute is on the 24-hour clock (ex : 2014-1-15T01:30). 
        """

        from datetime import datetime
        import pytz

        # Validate the Date formating
        try:
            snapshot_date = datetime.strptime(date, '%m-%d-%Y')
        except ValueError:
            sys.exit("Error: The date argument '{}' must be formatd as 'Month-Date-Year' (ex: 8-9-2018).".format(date))

        # Validate the Time formating
        try:
            snapshot_time = datetime.strptime(time, '%I:%M %p')
        except ValueError:
            sys.exit("Error: The time argument '{}' must be formatd as 'Hour:Minute AM/PM' (ex: 2:57 AM).".format(time))

        self.log("_date_time_conversion: Getting the Rubrik cluster timezone.")
        cluster_summary = self.get('v1', '/cluster/me')
        cluster_timezone = cluster_summary['timezone']['timezone']

        self.log("_date_time_conversion: Converting the provided time to the 24-hour clock.")
        snapshot_time_24_hour_clock = datetime.strftime(snapshot_time, "%H:%M")

        self.log("_date_time_conversion: Creating a combined date/time variable.")
        snapshot_datetime = datetime.strptime('{} {}'.format(date, snapshot_time_24_hour_clock), '%m-%d-%Y %H:%M')

        # Add Timezone to snapshot_datetime Variable
        timezone = pytz.timezone(cluster_timezone)
        snapshot_datetime = timezone.localize(snapshot_datetime)

        self.log("_date_time_conversion: Converting the time to UTC.\n")
        utc_timezone = pytz.UTC
        snapshot_datetime = snapshot_datetime.astimezone(utc_timezone)

        return snapshot_datetime.strftime('%Y-%m-%dT%H:%M')

    def pause_snapshots(self, object_name, object_type, timeout=180):
        """Pause all snapshot activity for the provided object.

        Arguments:
            object_name {str} -- The name of the Rubrik object to pause snapshots for.
            object_type {str} -- The Rubrik object type you wish to pause snaphots on. (choices: {vmware})

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster. (default: {180})

        Returns:
            str -- No change required. The '`object_type`' '`object_name`' is already paused.
            dict -- The full API response for `PATCH /v1/vmware/vm/{vm_id}`.
        """

        valid_object_type = ['vmware']

        if object_type not in valid_object_type:
            sys.exit("Error: The pause_snapshots() object_type argument must be one of the following: {}.".format(valid_object_type))

        if object_type == 'vmware':

            self.log("pause_snapshots: Searching the Rubrik cluster for the vSphere VM '{}'.".format(object_name))
            vm_id = self.object_id(object_name, object_type)

            self.log("pause_snapshots: Determing the current pause state of the vSphere VM '{}'.".format(object_name))
            api_request = self.get('v1', '/vmware/vm/{}'.format(vm_id))

            if api_request['blackoutWindowStatus']['isSnappableBlackoutActive'] is True:
                return "No change required. The '{}' '{}' is already paused.".format(object_type, object_name)
            else:
                self.log("pause_snapshots: Pausing Snaphots for the vSphere VM '{}'.".format(object_name))

                config = {}
                config['isVmPaused'] = True

                return self.patch('v1', '/vmware/vm/{}'.format(vm_id), config, timeout)

    def resume_snapshots(self, object_name, object_type, timeout=180):
        """Resume all snapshot activity for the provided object.

        Arguments:
            object_name {str} -- The name of the Rubrik object to resume snapshots for.
            object_type {str} -- The Rubrik object type you wish to resume snaphots on. (choices: {vmware})

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster. (default: {180})

        Returns:
            str -- No change required. The 'object_type' object 'object_name' is currently not paused.
            dict -- The full response for `PATCH /v1/vmware/vm/{vm_id}`.
        """

        valid_object_type = ['vmware']

        if object_type not in valid_object_type:
            sys.exit("Error: The resume_snapshots() object_type argument must be one of the following: {}.".format(valid_object_type))

        if object_type == 'vmware':

            self.log("resume_snapshots: Searching the Rubrik cluster for the vSphere VM '{}'.".format(object_name))
            vm_id = self.object_id(object_name, object_type)

            self.log("resume_snapshots: Determing the current pause state of the vSphere VM '{}'.".format(object_name))
            api_request = self.get('v1', '/vmware/vm/{}'.format(vm_id))

            if api_request['blackoutWindowStatus']['isSnappableBlackoutActive'] is False:
                return "No change required. The '{}' object '{}' is currently not paused.".format(object_type, object_name)
            else:
                self.log("resume_snapshots: Resuming Snaphots for the vSphere VM '{}'.".format(object_name))

                config = {}
                config['isVmPaused'] = False

                return self.patch('v1', '/vmware/vm/{}'.format(vm_id), config, timeout)



# Copyright 2018 Rubrik, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License prop
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Rubrik SDK Connect class.
"""

import base64
import requests
import sys
import os
import logging
from random import choice
import time

from .api import Api
from .cluster import Cluster
from .data_management import Data_Management
from .physical import Physical
from .cloud import Cloud


_CLUSTER = Cluster
_DATA_MANAGEMENT = Data_Management
_PHYSICAL = Physical
_API = Api
_CLOUD = Cloud


class Connect(_CLUSTER, _DATA_MANAGEMENT, _PHYSICAL, _CLOUD):
    """This class acts as the base class for the Rubrik SDK and serves as the main interaction point
    for its end users. It also contains various helper functions used throughout the SDK.

    Arguments:
        _CLUSTER {class} -- This class contains methods related to the managment of the Rubrik Cluster itself.
        _DATA_MANAGEMENT {class} - This class contains methods related to backup and restore operations for the various objects managed by the Rubrik Cluster.
        _PHYSICAL {class} - This class contains methods related to the managment of the Physical objects in the Rubrik Cluster.
    """

    def __init__(self, node_ip=None, username=None, password=None, enable_logging=False):
        """Constructor for the Connect class which is used to initialize the class variables.

        Keyword Arguments:
            node_ip {str} -- The Hostname or IP Address of a node in the Rubrik Cluster you wish to connect to. If a value is not provided we will check for a `rubrik_cdm_node_ip` environment variable. (default: {None})
            username {str} -- The Username you wish to use to connect to the Rubrik Cluster.. If a value is not provided we will check for a `rubrik_cdm_username` environment variable. (default: {None})
            password {str} -- The Password you wish to use to connect to the Rubrik Cluster.. If a value is not provided we will check for a `rubrik_cdm_password` environment variable. (default: {None})
            enable_logging {bool} -- Flag to determine if logging will be enabled for the SDK. (default: {False})
        """

        if node_ip is None:
            node_ip = os.environ.get('rubrik_cdm_node_ip')
            if node_ip is None:
                sys.exit("Error: The Rubrik CDM Node IP has not been provided.")
            else:
                self.node_ip = node_ip
        else:
            self.node_ip = node_ip

        if username is None:
            username = os.environ.get('rubrik_cdm_username')
            if username is None:
                sys.exit("Error: The Rubrik CDM Username has not been provided.")
            else:
                self.username = username
        else:
            self.username = username

        if password is None:
            password = os.environ.get('rubrik_cdm_password')
            if password is None:
                sys.exit("Error: The Rubrik CDM Password has not been provided.")
            else:
                self.password = password
        else:
            self.password = password

        if enable_logging is True:
            logging.getLogger().setLevel(logging.DEBUG)

        self.log("User Provided Node IP: {}".format(self.node_ip))
        self.log("Username: {}".format(self.username))
        self.log("Password: *******\n")

        self.node_ip = self.cluster_node_ip(timeout=5)

    @staticmethod
    def log(log_message):
        """Create properly formatted debug log messages.

        Arguments:
            log_message {str} -- The message to pass to the debug log.
        """
        log = logging.getLogger(__name__)
        log.debug(log_message)

    def _authorization_header(self):
        """Internal method used to create the authorization header used in the API calls.

        Returns:
            dict -- The authorization header that utilizes Basic authentication.
        """

        credentials = '{}:{}'.format(self.username, self.password)

        # Encode the Username:Password as base64
        authorization = base64.b64encode(credentials.encode())
        # Convert to String for API Call
        authorization = authorization.decode()

        authorization_header = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': 'Basic ' + authorization
        }

        return authorization_header

    @staticmethod
    def _header():
        """Internal method used to create the a header without authorization used in the API calls.

        Returns:
            dict -- The header that does not include any authorization.
        """

        header = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }

        return header

    @staticmethod
    def _api_validation(api_version, api_endpoint):
        """Internal method used to validate the API Version and API Endpoint provided by the end user

        Arguments:
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).
        """

        valid_api_versions = ['v1', 'internal']

        # Validate the API Version
        if api_version not in valid_api_versions:
            sys.exit("Error: Enter a valid API version {}.".format(valid_api_versions))

        # Validate the API Endpoint Syntax
        if type(api_endpoint) is not str:
            sys.exit("Error: The API Endpoint must be a string.")
        elif api_endpoint[0] != "/":
            sys.exit("Error: The API Endpoint should begin with '/'. (ex: /cluster/me)")
        elif api_endpoint[-1] == "/":
            sys.exit("Error: The API Endpoint should not end with '/'. (ex. /cluster/me)")


class Bootstrap(_API):
    """This class contains all functions related to the Bootstrapping of a Rubrik Cluster.

    Arguments:
        _API {class} - This class contains the base API methods that can be called independently or internally in standalone functions.
    """

    def __init__(self, node_ip, enable_logging=False):
        """Constructor for the Bootstrap class which is used to initialize the class variables.
        """

        if enable_logging is True:
            logging.getLogger().setLevel(logging.DEBUG)

        self.node_ip = node_ip
        self.log("User Provided Node IP: {}".format(self.node_ip))

        node_ip = [self.node_ip]

    def setup_cluster(self, cluster_name, admin_email, admin_password, management_gateway, management_subnet_mask, node_config=None, enable_encryption=True, dns_search_domains=None, dns_nameservers=None, ntp_servers=None, wait_for_completion=True, timeout=30):
        """Issues a bootstrap request to a specified Rubrik cluster

        Arguments:
            cluster_name {str} -- Unique name to assign to the Rubrik cluster.
            admin_email {str} -- The Rubrik cluster sends messages for the admin account to this email address.
            admin_password {str} --  Password for the admin account.
            management_gateway {str} --  IP address assigned to the management network gateway
            management_subnet_mask {str} -- Subnet mask assigned to the management network.

        Keyword Arguments:
            node_config {dict} -- The Node Name and IP formatted as a dictionary. (default: {None})
            enable_encryption {bool} -- Enable software data encryption at rest. When bootstraping a Cloud Cluster this value needs to be False. (default: {True})
            dns_search_domains {str} -- The search domain that the DNS Service will use to resolve hostnames that are not fully qualified. (default: {None})
            dns_nameservers {list} -- IPv4 addresses of DNS servers. (default: {['8.8.8.8']})
            ntp_servers {list} -- FQDN or IPv4 address of a network time protocol (NTP) server. (default: {['pool.ntp.org']})
            wait_for_completion {bool} -- Flag to determine if the function should wait for the bootstrap process to complete. (default: {True})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {30})

        Returns:
            dict -- The response returned by `POST /internal/cluster/me/bootstrap`.
        """

        if node_config is None or isinstance(node_config, dict) is not True:
            sys.exit('Error: You must provide a valid dictionary for "node_config".')

        if dns_search_domains is None:
            dns_search_domains = []
        elif isinstance(dns_search_domains, list) is not True:
            sys.exit('Error: You must provide a valid list for "dns_search_domains".')

        if dns_nameservers is None:
            dns_nameservers = ['8.8.8.8']
        elif isinstance(dns_nameservers, list) is not True:
            sys.exit('Error: You must provide a valid list for "dns_nameservers".')

        if ntp_servers is None:
            ntp_servers = ['pool.ntp.org']
        elif isinstance(ntp_servers, list) is not True:
            sys.exit('Error: You must provide a valid list for "ntp_servers".')

        bootstrap_config = {}
        bootstrap_config["enableSoftwareEncryptionAtRest"] = enable_encryption
        bootstrap_config["name"] = cluster_name
        bootstrap_config["dnsNameservers"] = dns_nameservers
        bootstrap_config["dnsSearchDomains"] = dns_search_domains
        bootstrap_config["ntpServers"] = ntp_servers

        bootstrap_config["adminUserInfo"] = {}
        bootstrap_config["adminUserInfo"]['password'] = admin_password
        bootstrap_config["adminUserInfo"]['emailAddress'] = admin_email
        bootstrap_config["adminUserInfo"]['id'] = "admin"

        bootstrap_config["nodeConfigs"] = {}
        for node_name, node_ip in node_config.items():
            bootstrap_config["nodeConfigs"][node_name] = {}
            bootstrap_config["nodeConfigs"][node_name]['managementIpConfig'] = {}
            bootstrap_config["nodeConfigs"][node_name]['managementIpConfig']['netmask'] = management_subnet_mask
            bootstrap_config["nodeConfigs"][node_name]['managementIpConfig']['gateway'] = management_gateway
            bootstrap_config["nodeConfigs"][node_name]['managementIpConfig']['address'] = node_ip

        while True:

            try:
                self.log('bootstrap: Starting the bootstrap process.')
                number_of_attempts = 1
                api_request = self.post('internal', '/cluster/me/bootstrap',
                                        bootstrap_config, timeout, authentication=False)
                break
            except SystemExit as bootstrap_error:
                if "Failed to establish a new connection: [Errno 111] Connection refused" in str(bootstrap_error):
                    self.log('bootstrap: Connection refused. Waiting 30 seconds for the node to initialize before trying again.')
                    number_of_attempts += 1
                    time.sleep(30)
                elif "Error: Cannot bootstrap from an already bootstrapped node" in str(bootstrap_error):
                    return "No change required. The Rubrik cluster is already bootstrapped."
                else:
                    self.log('bootstrap: Connection refused.')
                    sys.exit(bootstrap_error)

            if number_of_attempts == 12:
                sys.exit("Error: Unable to establish a connection to the Rubrik cluster.")

        request_id = api_request['id']

        if wait_for_completion == True:
            self.log('bootstrap: Waiting for the bootstrap process to complete.')
            while True:
                status = self.status(request_id)

                if status['status'] == 'IN_PROGRESS':
                    self.log("bootstrap_status: {}\n".format(status))
                    time.sleep(30)
                    continue
                elif status['status'] == 'FAILURE' or status['status'] == "FAILED":
                    sys.exit("Error: {}".format(status['message']))
                else:
                    self.log("{}".format(status))
                    return status

        return api_request

    def status(self, request_id="1", timeout=15):
        """Retrieves status of in progress bootstrap requests

        Keyword Arguments:
            request_id {str} -- ID of the bootstrap request(default: {"1"})
            timeout {int} -- The response timeout value, in seconds, of the API call. (default: {15})

        Returns:
            dict -- The response returned by `GET /internal/cluster/me/bootstrap?request_id={request_id}`.
        """

        self.log('status: Getting the status of the Rubrik Cluster bootstrap.')
        bootstrap_status_api_endpoint = '/cluster/me/bootstrap?request_id={}'.format(request_id)
        api_request = self.get('internal', bootstrap_status_api_endpoint, timeout=timeout, authentication=False)

        return api_request

    @staticmethod
    def log(log_message):
        """Create properly formatted debug log messages.

        Arguments:
            log_message {str} - - The message to pass to the debug log.
        """
        log = logging.getLogger(__name__)
        log.debug(log_message)

    @staticmethod
    def _header():
        """Internal method used to create the a header without authorization used in the API calls.

        Returns:
            dict - - The header that does not include any authorization.
        """

        header = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }

        return header

    @staticmethod
    def _api_validation(api_version, api_endpoint):
        """Internal method used to validate the API Version and API Endpoint provided by the end user

        Arguments:
            api_version {str} - - The version of the Rubrik CDM API to call.
            api_endpoint {str} - - The endpoint(ex. cluster / me) of the Rubrik CDM API to call.
        """

        valid_api_versions = ['v1', 'internal']

        # Validate the API Version
        if api_version not in valid_api_versions:
            sys.exit("Error: Enter a valid API version {}.".format(valid_api_versions))

        # Validate the API Endpoint Syntax
        if type(api_endpoint) is not str:
            sys.exit("Error: The API Endpoint must be a string.")
        elif api_endpoint[0] != "/":
            sys.exit("Error: The API Endpoint should begin with '/'. (ex: /cluster/me)")
        elif api_endpoint[-1] == "/":
            sys.exit("Error: The API Endpoint should not end with '/'. (ex. /cluster/me)")



# Copyright 2018 Rubrik, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License prop
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Rubrik SDK API class.
"""

import requests
import json
import sys
import time
try:
    from urllib import quote  # Python 2.X
except ImportError:
    from urllib.parse import quote  # Python 3+
from random import choice


class Api():
    """This class contains the base API methods that can be called independently or internally in standalone functions."""

    def __init__(self, node_ip):
        super().__init__(node_ip)

    def _common_api(self, call_type, api_version, api_endpoint, config=None, job_status_url=None, timeout=15, authentication=True):
        """Internal method that consolidates the base API functions.

        Arguments:
            call_type {str} -- The HTTP Method for the type of RESTful API call being made. (choices: {'GET', 'POST', 'PATCH', 'DELETE', and 'JOB_STATUS'.})
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, internal}) (choices: {v1, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).

        Keyword Arguments:
            config {dict} -- The specified data to send with `POST` and `PATCH` API calls. (default: {None})
            job_status_url {str} -- The job status URL provided by a previous API call. (default: {None})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})
            authentication {bool} -- Flag that specifies whether or not to utilize authentication when making the API call. (default: {True})

        Returns:
            dict -- The full API call response for the provided endpoint.
        """
        if call_type is not 'JOB_STATUS':
            self._api_validation(api_version, api_endpoint)

        """In order to dynamically select a node to interact with, the SDK will first use the node IP provided
        by the end user to get a list of all node IPs in the Rubrik cluster. This code will determine if the SDK has gathered
        that list yet and if it has will randomly select a node IP from the list to interact with."""
        if isinstance(self.node_ip, str):
            node_ip = self.node_ip
        else:
            node_ip = choice(self.node_ip)

        # Determine if authentication should be sent as part of the API Header
        if authentication == True:
            header = self._authorization_header()
        elif authentication == False:
            header = self._header()
        else:
            sys.exit('Error: "authentication" must be either True or False')

        try:
            # Determine which call type is being used and then set the relevant variables for that call type
            if call_type is 'GET':
                request_url = "https://{}/api/{}{}".format(node_ip, api_version, api_endpoint)
                request_url = quote(request_url, '://?=&')
                self.log('GET {}'.format(request_url))
                api_request = requests.get(request_url, verify=False, headers=header, timeout=timeout)
            elif call_type is 'POST':
                config = json.dumps(config)
                request_url = "https://{}/api/{}{}".format(node_ip, api_version, api_endpoint)
                self.log('POST {}'.format(request_url))
                self.log('Config: {}'.format(config))
                api_request = requests.post(request_url, verify=False, headers=header, data=config, timeout=timeout)
            elif call_type is 'PATCH':
                config = json.dumps(config)
                request_url = "https://{}/api/{}{}".format(node_ip, api_version, api_endpoint)
                self.log('PATCH {}'.format(request_url))
                self.log('Config: {}'.format(config))
                api_request = requests.patch(request_url, verify=False, headers=header, data=config, timeout=timeout)
            elif call_type is 'DELETE':
                request_url = "https://{}/api/{}{}".format(node_ip, api_version, api_endpoint)
                self.log('DELETE {}'.format(request_url))
                api_request = requests.delete(request_url, verify=False, headers=header, timeout=timeout)
            elif call_type is 'JOB_STATUS':
                self.log('JOB STATUS for {}'.format(job_status_url))
                api_request = requests.get(job_status_url, verify=False, headers=header, timeout=timeout)
            else:
                sys.exit('Error: the _common_api() call_type must be one of the following: {}'.format(
                    ['GET', 'POST', 'PATCH', 'DELETE', 'JOB_STATUS']))

            self.log(str(api_request) + "\n")
            try:
                api_response = api_request.json()
                # Check to see if an error message has been provided by Rubrik
                for key, value in api_response.items():
                    if key == "errorType" or key == 'message':
                        error_message = api_response['message']
                        api_request.raise_for_status()

            except:
                api_request.raise_for_status()
        except requests.exceptions.ConnectTimeout:
            sys.exit('Error: Unable to establish a connection to the Rubrik cluster.')
        except requests.exceptions.RequestException as error:
            # If "error_message" has be defined sys.exit that message else sys.exit the request exception error
            try:
                error_message
            except NameError:
                sys.exit(error)
            else:
                sys.exit('Error: ' + error_message)
        else:
            try:
                return api_request.json()
            except:
                return {'status_code': api_request.status_code}

    def get(self, api_version, api_endpoint, timeout=15, authentication=True):
        """Send a GET request to the provided Rubrik API endpoint.

        Arguments:
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})
            authentication {bool} -- Flag that specifies whether or not to utilize authentication when making the API call. (default: {True})

        Returns:
            dict -- The response body of the API call.
        """

        return self._common_api('GET', api_version, api_endpoint, config=None,
                                job_status_url=None, timeout=timeout, authentication=authentication)

    def post(self, api_version, api_endpoint, config, timeout=15, authentication=True):
        """Send a POST request to the provided Rubrik API endpoint.

        Arguments:
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).
            config {dict} -- The specified data to send with the API call.

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})
            authentication {bool} -- Flag that specifies whether or not to utilize authentication when making the API call. (default: {True})

        Returns:
            dict -- The response body of the API call.
        """

        return self._common_api('POST', api_version, api_endpoint, config=config,
                                job_status_url=None, timeout=timeout, authentication=authentication)

    def patch(self, api_version, api_endpoint, config, timeout=15, authentication=True):
        """Send a PATCH request to the provided Rubrik API endpoint.

        Arguments:
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).
            config {dict} -- The specified data to send with the API call.

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})
            authentication {bool} -- Flag that specifies whether or not to utilize authentication when making the API call. (default: {True})

        Returns:
            dict -- The response body of the API call.
        """

        return self._common_api('PATCH', api_version, api_endpoint, config=config,
                                job_status_url=None, timeout=timeout, authentication=authentication)

    def delete(self, api_version, api_endpoint, timeout=15, authentication=True):
        """Send a DELETE request to the provided Rubrik API endpoint.

        Arguments:
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})
            authentication {bool} -- Flag that specifies whether or not to utilize authentication when making the API call. (default: {True})

        Returns:
            dict -- The response body of the API call.
        """

        return self._common_api('DELETE', api_version, api_endpoint, config=None,
                                job_status_url=None, timeout=timeout, authentication=authentication)

    def job_status(self, url, wait_for_completion=True, timeout=15):
        """Certain Rubrik operations (on-demand snapshots, live mounts, etc.) may not complete instantaneously. In those cases we have the ability to monitor the status of the
        job through a job status url provided in the actions API response body. This function will perform a GET operation on the provided url
        and return the jobs status.

        Arguments:
            url {str} -- The job status URL provided by a previous API call.

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})
            wait_for_completion {bool} -- Flag that determines if the method should wait for the job to complete before exiting. (default: {True})

        Returns:
            dict -- The response body of the API call.
        """

        if not isinstance(wait_for_completion, bool):
            sys.exit('Error: The job_status() wait_for_completion argument must be True or False.')

        if wait_for_completion:
            self.log('Job Status: Waiting for the job to complete.')
            api_call = self._common_api('JOB_STATUS', api_version=None, api_endpoint=None,
                                        config=None, job_status_url=url, timeout=timeout)

            while True:

                api_call = self._common_api('JOB_STATUS', api_version=None, api_endpoint=None,
                                            config=None, job_status_url=url, timeout=timeout)

                job_status = api_call['status']

                if job_status == "SUCCEEDED":
                    self.log('Job Progress 100%\n')
                    job_status = api_call['status']
                    break
                elif job_status == "QUEUED" or "RUNNING":
                    self.log('Job Progress {}%\n'.format(api_call['progress']))
                    job_status = api_call['status']
                    time.sleep(10)
                    continue
                else:
                    sys.exit('Error: {}'.format(str(api_call)))

        else:
            api_call = self._common_api('JOB_STATUS', api_version=None, api_endpoint=None,
                                        config=None, job_status_url=url, timeout=timeout)

        return api_call



# Copyright 2018 Rubrik, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License prop
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Rubrik SDK Physical class.
"""

import sys
from .api import Api

_API = Api


class Physical(_API):
    """This class contains methods related to the managment of the Physical objects in the Rubrik cluster."""

    def add_physical_host(self, hostname, timeout=60):
        """Add a physical host to the Rubrik cluster.

        Arguments:
            hostname {str} -- The hostname or IP Address of the physical host you want to add to the Rubrik cluster.

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {60})

        Returns:
            str -- No change requird. The host '`hostname`' is already connected to the Rubrik cluster.
            dict -- The full API response for `POST /v1'/host'`.
        """

        self.log('Searching the Rubrik cluster for the current hosts.')
        current_hosts = self.get('v1', '/host')

        for host in current_hosts['data']:
            if host['hostname'] == hostname:
                return "No change requird. The host '{}' is already connected to the Rubrik cluster.".format(hostname)

        config = {}
        config['hostname'] = hostname
        config['hasAgent'] = True

        self.log("Adding the host '{}' to the Rubrik cluster.".format(hostname))
        return self.post('v1', '/host', config, timeout)

    def delete_physical_host(self, hostname, timeout=120):
        """Delete a physical host from the Rubrik cluster.

        Arguments:
            hostname {str} -- The hostname or IP Address of the physical host you wish to remove from the Rubrik cluster.

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {120})

        Returns:
            str -- No change required. The host '`hostname`' is not connected to the Rubrik cluster.
            dict -- The full API response for `DELETE /v1'/host/{host_id}`.
        """

        self.log('Searching the Rubrik cluster for the current hosts.')
        current_hosts = self.get('v1', '/host')

        host_present = False

        for host in current_hosts['data']:
            if host['hostname'] == hostname:
                host_present = True
                host_id = host['id']
                break

        if host_present is False:
            return "No change required. The host '{}' is not connected to the Rubrik cluster.".format(hostname)

        self.log("Deleting the host '{}' from the Rubrik cluster.".format(hostname))
        return self.delete('v1', '/host/{}'.format(host_id), timeout)

    def create_physical_fileset(self, name, operating_system, include, exclude, exclude_exception, follow_network_shares=False, backup_hidden_folders=False, timeout=15):
        """Create a Fileset for a Linux or Windows machine.

        Arguments:
            name {str} -- The name of the Fileset you wish to create.
            operating_system {str} -- The operating system type of the Fileset you are creating. (choices: {Linux, Windows.})
            include {list} -- The full paths or wildcards that define the objects to include in the Fileset backup (ex: ['/usr/local', '*.pdf']).
            exclude {list} -- The full paths or wildcards that define the objects to exclude from the Fileset backup (ex: ['/user/local/temp', '*.mov', '*.mp3']).
            exclude_exception {list} -- The full paths or wildcards that define the objects that are exempt from the `excludes` variables. (ex. ['/company/*.mp4').

        Keyword Arguments:
            follow_network_shares {bool} -- Include or exclude locally-mounted remote file systems from backups. (default: {False}) 
            backup_hidden_folders {bool} -- Include or exclude hidden folders inside locally-mounted remote file systems from backups. (default: {False})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})

        Returns:
            str -- No change required. The Rubrik cluster already has a `operating_system` Fileset named '`name`' configured with the provided variables.
            dict -- The full response for the `POST /internal/fileset_template/bulk` API endpoint.
        """

        valid_operating_system = ['Linux', 'Windows']

        if operating_system not in valid_operating_system:
            sys.exit("Error: The create_physical_fileset() operating_system argument must be one of the following: {}.".format(
                valid_operating_system))

        if isinstance(follow_network_shares, bool) is False:
            sys.exit("Error: The 'follow_network_shares' argument must be True or False.")
        elif isinstance(backup_hidden_folders, bool) is False:
            sys.exit("Error: The 'backup_hidden_folders' argument must be True or False.")
        elif isinstance(include, list) is False:
            sys.exit("Error: The 'include' argument must be a list object.")
        elif isinstance(exclude, list) is False:
            sys.exit("Error: The 'exclude' argument must be a list object.")
        elif isinstance(exclude_exception, list) is False:
            sys.exit("Error: The 'exclude_exception' argument must be a list object.")

        config = {}
        config['name'] = name
        config['includes'] = sorted(include)
        config['excludes'] = sorted(exclude)
        config['exceptions'] = sorted(exclude_exception)
        config['allowBackupHiddenFoldersInNetworkMounts'] = backup_hidden_folders
        config['allowBackupNetworkMounts'] = follow_network_shares
        config['operatingSystemType'] = operating_system

        self.log("create_fileset: Searching the Rubrik cluster for all current {} Filesets.".format(operating_system))
        current_filesets = self.get(
            'v1', '/fileset_template?primary_cluster_id=local&operating_system_type={}&name={}'.format(operating_system, name), timeout)

        current_config = {}
        if current_filesets['data']:
            current_config['name'] = current_filesets['data'][0]['name']
            current_config['includes'] = sorted(current_filesets['data'][0]['includes'])
            current_config['excludes'] = sorted(current_filesets['data'][0]['excludes'])
            current_config['exceptions'] = sorted(current_filesets['data'][0]['exceptions'])
            current_config['allowBackupHiddenFoldersInNetworkMounts'] = current_filesets['data'][0]['allowBackupHiddenFoldersInNetworkMounts']
            current_config['operatingSystemType'] = current_filesets['data'][0]['operatingSystemType']
            current_config['allowBackupNetworkMounts'] = current_filesets['data'][0]['allowBackupNetworkMounts']

        if current_config == config:
            return "No change required. The Rubrik cluster already has a {} Fileset named '{}' configured with the provided variables.".format(operating_system, name)

        # Add the config dict to a list
        model = []
        model.append(config)

        self.log("create_fileset: Creating the '{}' Fileset.".format(name))
        return self.post('internal', '/fileset_template/bulk', model, timeout=timeout)

    def create_nas_fileset(self, name, share_type, include, exclude, exclude_exception, follow_network_shares=False, timeout=15):
        """Create a NAS Fileset.

        Arguments:
            name {str} -- The name of the Fileset you wish to create.
            share_type {str} -- The type of NAS Share you wish to backup. (choices: {NFS, SMB})
            include {list} -- The full paths or wildcards that define the objects to include in the Fileset backup.
            exclude {list} -- The full paths or wildcards that define the objects to exclude from the Fileset backup.
            exclude_exception {list} -- The full paths or wildcards that define the objects that are exempt from the `excludes` variables.

        Keyword Arguments:
            follow_network_shares {bool} -- Include or exclude locally-mounted remote file systems from backups. (default: {False}) 
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})

        Returns:
            str -- No change required. The Rubrik cluster already has a NAS Fileset named '`name`' configured with the provided variables."
            dict -- The full response for the `POST /internal/fileset_template/bulk` API endpoint.
        """

        valid_share_type = ['NFS', 'SMB']

        if share_type not in valid_share_type:
            sys.exit("Error: The create_fileset() share_type argument must be one of the following: {}.".format(valid_share_type))

        if isinstance(follow_network_shares, bool) is False:
            sys.exit("Error: The 'follow_network_shares' argument must be True or False.")
        elif isinstance(include, list) is False:
            sys.exit("Error: The 'include' argument must be a list object.")
        elif isinstance(exclude, list) is False:
            sys.exit("Error: The 'exclude' argument must be a list object.")
        elif isinstance(exclude_exception, list) is False:
            sys.exit("Error: The 'exclude_exception' argument must be a list object.")

        config = {}
        config['name'] = name
        config['includes'] = sorted(include)
        config['excludes'] = sorted(exclude)
        config['exceptions'] = sorted(exclude_exception)
        config['allowBackupHiddenFoldersInNetworkMounts'] = follow_network_shares
        config['shareType'] = share_type

        self.log("create_fileset: Searching the Rubrik cluster for all current NAS Filesets.")
        current_filesets = self.get(
            'v1', '/fileset_template?primary_cluster_id=local&operating_system_type=NONE&name={}'.format(name), timeout=timeout)

        current_config = {}
        if current_filesets['data']:
            current_config['name'] = current_filesets['data'][0]['name']
            current_config['includes'] = sorted(current_filesets['data'][0]['includes'])
            current_config['excludes'] = sorted(current_filesets['data'][0]['excludes'])
            current_config['exceptions'] = sorted(current_filesets['data'][0]['exceptions'])
            current_config['allowBackupHiddenFoldersInNetworkMounts'] = current_filesets['data'][0]['allowBackupHiddenFoldersInNetworkMounts']
            current_config['shareType'] = current_filesets['data'][0]['shareType']

        if current_config == config:
            return "No change required. The Rubrik cluster already has a NAS Fileset named '{}' configured with the provided variables.".format(name)

        # Add the config dict to a list
        model = []
        model.append(config)

        self.log("create_fileset: Creating the '{}' Fileset.".format(name))
        return self.post('internal', '/fileset_template/bulk', model, timeout=timeout)

    def assign_physical_host_fileset(self, hostname, fileset_name, operating_system, sla_name, include=None, exclude=None, exclude_exception=None, follow_network_shares=False, backup_hidden_folders=False, timeout=30):
        """Assign a Fileset to a Linux or Windows machine. If you have multiple Filesets with identical names, you will need to populate the Filesets properties (i.e this functions keyword arguments) 
        to find a specific match. Filesets with identical names and properties are not supported.

        Arguments:
            hostname {str} -- The hostname or IP Address of the physical host you wish to associate to the Fileset.
            fileset_name {str} -- The name of the Fileset you wish to assign to the Linux or Windows host.
            operating_system {str} -- The operating system of the physical host you are assigning a Fileset to. (choices: {Linux, Windows})
            sla_name {str} -- The name of the SLA Domain to associate with the Fileset.

        Keyword Arguments:
            include {list} -- The full paths or wildcards that define the objects to include in the Fileset backup (ex: ['/usr/local', '*.pdf']). (default: {None})
            exclude {list} -- The full paths or wildcards that define the objects to exclude from the Fileset backup (ex: ['/user/local/temp', '*.mov', '*.mp3']). (default: {None})
            exclude_exception {list} -- The full paths or wildcards that define the objects that are exempt from the `excludes` variables. (ex: ['/company/*.mp4']). (default: {None})
            follow_network_shares {bool} -- Include or exclude locally-mounted remote file systems from backups. (default: {False})
            backup_hidden_folders {bool} -- Include or exclude hidden folders inside locally-mounted remote file systems from backups. (default: {False})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {30})

        Returns:
            str -- No change required. The `operating_system` Fileset '`fileset_name`' is already assigned to the SLA Domain '`sla_name`' on the physical host '`hostname`'.
            tuple -- When a new Fileset is created the following tuple will be returned: (Full API response from `POST /v1/fileset`, Full API response from `POST /v1/fileset/{id}`)
            dict -- When the Fileset already exsits but is assigned to the wrong the SLA the Full API response from `POST `v1/fileset/{id}` is returned.
        """

        valid_operating_system = ['Linux', 'Windows']

        if operating_system not in valid_operating_system:
            sys.exit("Error: The create_physical_fileset() operating_system argument must be one of the following: {}.".format(
                valid_operating_system))

        if include == None:
            include = []

        if exclude == None:
            exclude = []

        if exclude_exception == None:
            exclude_exception = []

        if isinstance(follow_network_shares, bool) is False:
            sys.exit("Error: The 'follow_network_shares' argument must be True or False.")
        elif isinstance(backup_hidden_folders, bool) is False:
            sys.exit("Error: The 'backup_hidden_folders' argument must be True or False.")
        elif isinstance(include, list) is False:
            sys.exit("Error: The 'include' argument must be a list object.")
        elif isinstance(exclude, list) is False:
            sys.exit("Error: The 'exclude' argument must be a list object.")
        elif isinstance(exclude_exception, list) is False:
            sys.exit("Error: The 'exclude_exception' argument must be a list object.")

        self.log("assign_physical_host_fileset: Searching the Rubrik cluster for the {} physical host {}.".format(
            operating_system, hostname))
        current_hosts = self.get(
            'v1', '/host?operating_system_type={}&primary_cluster_id=local&hostname={}'.format(operating_system, hostname), timeout)

        if current_hosts['total'] >= 1:
            for host in current_hosts['data']:
                if host['hostname'] == hostname:
                    host_id = host['id']
                    break
        try:
            host_id
        except NameError:
            sys.exit("Error: The Rubrik cluster is not connected to a {} physical host named '{}'.".format(
                operating_system, hostname))

        self.log("assign_physical_host_fileset: Searching the Rubrik cluster for all current {} Filesets.".format(
            operating_system))
        current_filesets_templates = self.get('v1', '/fileset_template?primary_cluster_id=local&operating_system_type={}&name={}'.format(
            operating_system, fileset_name), timeout)

        number_of_matches = 0
        if current_filesets_templates['total'] == 0:
            sys.exit("Error: The Rubrik cluster does not have a {} Fileset named '{}'.".format(operating_system, fileset_name))
        elif current_filesets_templates['total'] > 1:
            for fileset_template in current_filesets_templates['data']:
                if fileset_template['name'] == fileset_name:
                    number_of_matches += 1

            if number_of_matches > 1:
                # If there are multiple Filesets with the same name us all of the possible config values to try and find the correct Fileset
                number_of_matches = 0
                for fileset_template in current_filesets_templates['data']:
                    if fileset_template['name'] == fileset_name \
                            and fileset_template['includes'] == include \
                            and fileset_template['excludes'] == exclude \
                            and fileset_template['exceptions'] == exclude_exception \
                            and fileset_template['allowBackupHiddenFoldersInNetworkMounts'] == follow_network_shares \
                            and fileset_template['allowBackupNetworkMounts'] == backup_hidden_folders:
                        fileset_template_id = fileset_template['id']
                        number_of_matches = 1

                try:
                    fileset_template_id
                except NameError:
                    if number_of_matches > 0:
                        # If no unique matches are found provide an error message
                        if include != [] \
                                or exclude != [] \
                                or exclude_exception != [] \
                                or follow_network_shares != False \
                                or backup_hidden_folders != False:
                            # Error message that first checks to see if any of the extra variables are populated with anything besides the default (aka the user tried to be as unique as possible)
                            sys.exit("Error: The Rubrik cluster contains multiple {} Filesets named '{}' that match all of the populate function arguments. Please use a unique Fileset.".format(
                                operating_system, fileset_name))
                        else:
                            sys.exit("Error: The Rubrik cluster contains multiple {} Filesets named '{}'. Please populate all function arguments to find a more specific match.".format(
                                operating_system, fileset_name))
                    sys.exit("Error: The Rubrik cluster contains multiple {} Filesets named '{}'. Please populate all function arguments to find a more specific match.".format(
                        operating_system, fileset_name))

        if current_filesets_templates['total'] == 1 or number_of_matches == 1:
            for fileset_temmplate in current_filesets_templates['data']:
                if fileset_temmplate['name'] == fileset_name:
                    fileset_template_id = fileset_temmplate['id']

        self.log("assign_physical_host_fileset: Searching the Rubrik cluster for the SLA Domain '{}'.".format(sla_name))
        sla_id = self.object_id(sla_name, 'sla')

        self.log("assign_physical_host_fileset: Getting the properties of the {} Fileset.".format(fileset_name))
        current_fileset = self.get(
            'v1', '/fileset?primary_cluster_id=local&host_id={}&is_relic=false&template_id={}'.format(host_id, fileset_template_id), timeout)

        if current_fileset['total'] == 0:
            self.log("assign_physical_host_fileset: Assigning the '{}' Fileset to the {} physical host '{}'.".format(
                fileset_name, operating_system, hostname))

            config = {}
            config['hostId'] = host_id
            config['templateId'] = fileset_template_id

            create_fileset = self.post('v1', '/fileset', config, timeout)
            fileset_id = create_fileset['id']

            config = {}
            config['configuredSlaDomainId'] = sla_id
            assign_sla = self.patch('v1', '/fileset/{}'.format(fileset_id), config, timeout)

            return (create_fileset, assign_sla)
        elif current_fileset['total'] == 1 and current_fileset['data'][0]['configuredSlaDomainId'] != sla_id:

            self.log("assign_physical_host_fileset: Assigning the '{}' SLA Domain to the '{}' Fileset attached to the {} physical host '{}'.".format(
                sla_name, fileset_name, operating_system, hostname))
            fileset_id = current_fileset['data'][0]['id']
            config = {}
            config['configuredSlaDomainId'] = sla_id
            return self.patch('v1', '/fileset/{}'.format(fileset_id), config, timeout)

        elif current_fileset['total'] == 1 and current_fileset['data'][0]['configuredSlaDomainId'] == sla_id:
            return "No change required. The {} Fileset '{}' is already assigned to the SLA Domain '{}' on the physical host '{}'.".format(operating_system, fileset_name, sla_name, hostname)



# Copyright 2018 Rubrik, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License prop
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Rubrik SDK Cloud class.
"""

import sys
import os
import re
from .api import Api

_API = Api


class Cloud(_API):
    """This class contains methods for the managment of Cloud related functionality on the Rubrik cluster."""

    def aws_s3_cloudout(self, aws_bucket_name, archive_name='default', aws_region=None, aws_access_key=None, aws_secret_key=None, kms_master_key_id=None, rsa_key=None, storage_class='standard', timeout=30):
        """Add a new AWS S3 archival location to the Rubrik cluster.

        Arguments:
            aws_bucket_name {str} -- The name of the AWS S3 bucket you wish to use as an archive target. The bucket name will automatically have all whitespace removed, all letters lowercased, and can not contain any of the following characters: `_\/*?%.:\|<>`.

        Keyword Arguments:
            aws_region {str} -- The name of the AWS region where the bucket is located. If set to the default `None` keyword argument, we will look for a `AWS_DEFAULT_REGION` environment variable to pull the value from. (default: {None}) (choices: {ap-south-1, ap-northeast-2, ap-southeast-1, ap-southeast-2, ap-northeast-1, ca-central-1, cn-north-1, cn-northwest-1, eu-central-1, eu-west-1, eu-west-2, eu-west-3, sa-east-1, us-gov-west-1, us-west-1, us-east-1, us-east-2, us-west-2})
            aws_access_key {str} -- The access key of a AWS account with the required permissions. If set to the default `None` keyword argument, we will look for a `AWS_ACCESS_KEY_ID` environment variable to pull the value from. (default: {None})
            aws_secret_key {str} -- The secret key of a AWS account with the required permissions. If set to the default `None` keyword argument, we will look for a `AWS_SECRET_ACCESS_KEY` environment variable to pull the value from. (default: {None})
            kms_master_key_id {str} -- The AWS KMS master key ID that will be used to encrypt the archive data. If set to the default `None` keyword argument, you will need to provide a `rsa_key` instead. (default: {None})
            rsa_key {str} -- The RSA key that will be used to encrypt the archive data. A key can be generated through `openssl genrsa -out rubrik_encryption_key.pem 2048`. If set to the default `None` keyword argument, you will need to provide a `kms_master_key_id` instead.  (default: {None})
            archive_name {str} -- The name of the archive location used in the Rubrik GUI. If set to 'default' the following naming convention will be used: "AWS:S3:`aws_bucket_name`" (default: {'default'})
            storage_class {str} -- The AWS storage class you wish to use. (default: {'standard'}) (choices: {standard, 'standard_ia, reduced_redundancy})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {30})


        Returns:
            str -- No change required. The '`name`' archival location is already configured on the Rubrik cluster.
            dict -- The full API response for `POST /internal/archive/object_store'`.
        """

        valid_aws_regions = ['ap-south-1', 'ap-northeast-2', 'ap-southeast-1', 'ap-southeast-2', 'ap-northeast-1', 'ca-central-1', 'cn-north-1',
                             'cn-northwest-1', 'eu-central-1', 'eu-west-1', 'eu-west-2', 'eu-west-3', 'sa-east-1', 'us-gov-west-1' 'us-west-1' 'us-east-1', 'us-east-2', 'us-west-2']

        valid_storage_classes = ['standard', 'standard_ia', 'reduced_redundancy']

        if re.compile(r'[_\/*?%.:|<>]').findall(aws_bucket_name):
            sys.exit("Error: The `aws_bucket_name` may not contain any of the following characters: _\/*?%.:|<>")

        if aws_region is None:
            aws_region = os.environ.get('AWS_DEFAULT_REGION')
            if aws_region is None:
                sys.exit("Error: `aws_region` has not been provided.")

        if aws_access_key is None:
            aws_access_key = os.environ.get('AWS_ACCESS_KEY_ID')
            if aws_access_key is None:
                sys.exit("Error: `aws_access_key` has not been provided.")

        if aws_secret_key is None:
            aws_secret_key = os.environ.get('AWS_SECRET_ACCESS_KEY')
            if aws_secret_key is None:
                sys.exit("Error: `aws_secret_key` has not been provided.")

        if aws_region not in valid_aws_regions:
            sys.exit('Error: The `aws_region` must be one of the following: {}'.format(valid_aws_regions))

        if storage_class not in valid_storage_classes:
            sys.exit('Error: The `storage_class` must be one of the following: {}'.format(valid_storage_classes))
        else:
            storage_class = storage_class.upper()

        if archive_name == 'default':
            archive_name = 'AWS:S3:{}'.format(aws_bucket_name.lower().strip())

        if kms_master_key_id is None and rsa_key is None:
            sys.exit("Error: You must populated either `kms_master_key_id` or `rsa_key`.")
        elif kms_master_key_id is not None and rsa_key is not None:
            sys.exit("Error: Both `kms_master_key_id` or `rsa_key` have been populated. You may only use one.")

        self.log("aws_s3_cloudout: Searching the Rubrik cluster for archival locations.")
        archives_on_cluster = self.get('internal', '/archive/object_store')

        config = {}
        config['name'] = archive_name
        config['bucket'] = aws_bucket_name.lower().strip()
        config['defaultRegion'] = aws_region
        config['storageClass'] = storage_class
        config['accessKey'] = aws_access_key
        config['secretKey'] = aws_secret_key
        if kms_master_key_id:
            config['kmsMasterKeyId'] = kms_master_key_id
        elif rsa_key:
            config['pemFileContent'] = rsa_key
        config['objectStoreType'] = 'S3'

        # Create a new dictionary that includes only the values returned by archives_on_cluster
        redacted_archive_definition = {}
        redacted_archive_definition['objectStoreType'] = 'S3'
        redacted_archive_definition['name'] = archive_name
        redacted_archive_definition['accessKey'] = aws_access_key
        redacted_archive_definition['bucket'] = aws_bucket_name.lower().strip()
        redacted_archive_definition['defaultRegion'] = aws_region
        redacted_archive_definition['storageClass'] = storage_class

        for archive in archives_on_cluster['data']:
            # If present, remove the Cloud On Configuration for comparison
            archive_definition = archive['definition']
            try:
                del archive_definition['defaultComputeNetworkConfig']
            except:
                pass
            if archive_definition == redacted_archive_definition:
                return "No change required. The '{}' archival location is already configured on the Rubrik cluster.".format(archive_name)
            if archive['definition']['objectStoreType'] == 'S3' and archive['definition']['name'] == archive_name:
                sys.exit(
                    "Error: Archival location with name '{}' already exists. Please enter a unique `archive_name`.".format(archive_name))

        self.log("aws_s3_cloudout: Creating the AWS S3 archive location.")
        return self.post('internal', '/archive/object_store', config, timeout)

    def aws_s3_cloudon(self, archive_name, vpc_id, subnet_id, security_group_id, timeout=30):
        """Enable CloudOn for an exsiting AWS S3 archival location.

        Arguments:
            archive_name {str} -- The name of the archive location used in the Rubrik GUI.

        Keyword Arguments:
            vpc_id {str} -- The AWS VPC ID used by Rubrik cluster to launch a temporary Rubrik instance in AWS for instantiation.
            subnet_id {str} -- The AWS Subnet ID used by Rubrik cluster to launch a temporary Rubrik instance in AWS for instantiation.
            security_group_id {str} -- The AWS Security Group ID used by Rubrik cluster to launch a temporary Rubrik instance in AWS for instantiation.
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {30})

        Returns:
            str -- No change required. The '`name`' archival location is already configured on the Rubrik cluster.
            dict -- The full API response for `PATCH /internal/archive/object_store/{id}`.
        """

        self.log("aws_s3_cloudon: Searching the Rubrik cluster for archival locations.")
        archives_on_cluster = self.get('internal', '/archive/object_store')

        config = {}
        config['defaultComputeNetworkConfig'] = {}
        config['defaultComputeNetworkConfig']['subnetId'] = subnet_id
        config['defaultComputeNetworkConfig']['vNetId'] = vpc_id
        config['defaultComputeNetworkConfig']['securityGroupId'] = security_group_id

        for archive in archives_on_cluster['data']:
            if archive['definition']['objectStoreType'] == 'S3' and archive['definition']['name'] == archive_name:
                # If present, remove the Cloud On configuration for proper comparison
                try:
                    if archive['definition']['defaultComputeNetworkConfig'] == config['defaultComputeNetworkConfig']:
                        return "No change required. The '{}' archival location is already configured for CloudOn.".format(archive_name)
                except KeyError:
                    self.log("aws_s3_cloudon: Updating the archive location for CloudOn.")
                    return self.patch('internal', "/archive/object_store/{}".format(archive['id']), config, timeout)

        sys.exit("Error: The Rubrik cluster does not have an archive location named '{}'.".format(archive_name))

    def azure_cloudout(self, container, azure_access_key, storage_account_name, rsa_key, archive_name='default', instance_type='default', timeout=30):
        """Add a new Azure archival location to the Rubrik cluster.

        Arguments:
            container {str} -- The name of the Azure storage container you wish to use as an archive. The container name will automatically be lowercased and can not contain any of the following characters: `_\/*?%.:\|<>`.
            azure_access_key {str} -- The access key for the Azure storage account.
            storage_account_name {str} -- The name of the Storage Account that the `container` belongs to.
            rsa_key {str} -- The RSA key that will be used to encrypt the archive data. A key can be generated through `openssl genrsa -out rubrik_encryption_key.pem 2048`.

        Keyword Arguments:
            archive_name {str} -- The name of the archive location used in the Rubrik GUI. If set to `default`, the following naming convention will be used: "Azure:`container`" (default: {'default'})
            instance_type {str} -- The Cloud Platform type of the archival location. (default: {'default'}) (choices: {'default', 'china', 'germany', 'government'})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {30})

        Returns:
            str -- No change required. The '`name`' archival location is already configured on the Rubrik cluster.
            dict -- The full API response for `POST /internal/archive/object_store`.
        """

        container = container.lower()

        if re.compile(r'[_\/*?%.:|<>]').findall(container):
            sys.exit("Error: The `container` may not contain any of the following characters: _\/*?%.:|<>")

        valid_instance_types = ['default', 'china', 'germany', 'government']

        if instance_type not in valid_instance_types:
            sys.exit('Error: The `instance_type` argument must be one of the following: {}'.format(valid_instance_types))

        if archive_name == 'default':
            archive_name = 'Azure:{}'.format(container)

        self.log("azure_cloudout: Searching the Rubrik cluster for archival locations.")
        archives_on_cluster = self.get('internal', '/archive/object_store')

        config = {}
        config['name'] = archive_name
        config['bucket'] = container
        config['accessKey'] = storage_account_name
        config['secretKey'] = azure_access_key
        config['pemFileContent'] = rsa_key
        config['objectStoreType'] = 'Azure'
        if instance_type == 'government':
            config['endpoint'] = 'core.usgovcloudapi.net'
        elif instance_type == 'germany':
            config['endpoint'] = 'core.cloudapi.de'
        elif instance_type == 'china':
            config['endpoint'] = 'core.chinacloudapi.cn'

        # Create a new dictionary that includes only the values returned by archives_on_cluster
        redacted_archive_definition = {}
        redacted_archive_definition['objectStoreType'] = 'Azure'
        redacted_archive_definition['name'] = archive_name
        redacted_archive_definition['accessKey'] = storage_account_name
        redacted_archive_definition['bucket'] = container
        if instance_type == 'government':
            redacted_archive_definition['endpoint'] = 'core.usgovcloudapi.net'
        elif instance_type == 'germany':
            redacted_archive_definition['endpoint'] = 'core.cloudapi.de'
        elif instance_type == 'china':
            redacted_archive_definition['endpoint'] = 'core.chinacloudapi.cn'

        for archive in archives_on_cluster['data']:
            # If present, remove the Cloud On Configuration for comparison
            archive_definition = archive['definition']
            try:
                del archive_definition['defaultComputeNetworkConfig']
                del archive_definition['isComputeEnabled']
                del archive_definition['azureComputeSummary']
            except:
                pass

            if archive_definition == redacted_archive_definition:
                return "No change required. The '{}' archival location is already configured on the Rubrik cluster.".format(archive_name)
            if archive['definition']['objectStoreType'] == 'Azure' and archive['definition']['name'] == archive_name:
                sys.exit("Error: Archival location with name '{}' already exists. Please enter a unique `name`.".format(archive_name))

        self.log("azure_cloudout: Creating the Azure archive location.")
        return self.post('internal', '/archive/object_store', config)

    def azure_cloudon(self, archive_name, container, storage_account_name, application_id, application_key, tenant_id, region, virtual_network_id, subnet_name, security_group_id, timeout=30):
        """Enable CloudOn for an exsiting AWS S3 archival location.

        Arguments:
            archive_name {str} -- The name of the archive location used in the Rubrik GUI.
            container {str} -- The name of the Azure storage container being used as the archive target. The container name will automatically be lowercased and can not contain any of the following characters: `_\/*?%.:\|<>`.
            storage_account_name {str} -- The name of the Storage Account that the `container` belongs to.
            application_id {str} -- The ID of the application registered in Azure Active Directory.
            application_key {str} -- The key of the application registered in Azure Active Directory.
            tenant_id {str} -- The tenant ID, also known as the directory ID, found under the Azure Active Directory properties.
            region {str} -- The name of the Azure region where the `container` is located. (choices: {westus, westus2, centralus, eastus, eastus2, northcentralus, southcentralus, westcentralus, canadacentral, canadaeast, brazilsouth, northeurope, westeurope, uksouth, ukwest, eastasia, southeastasia, japaneast, japanwest, australiaeast australiasoutheast, centralindia, southindia, westindia, koreacentral, koreasouth})
            virtual_network_id {str} -- The Azure virtual network ID used by Rubrik cluster to launch a temporary Rubrik instance in Azure for instantiation.
            subnet_name {str} -- The Azure subnet name used by Rubrik cluster to launch a temporary Rubrik instance in Azure for instantiation.
            security_group_id {str} -- The Azure Security Group ID used by Rubrik cluster to launch a temporary Rubrik instance in Azure for instantiation.

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {30})


        Returns:
            str -- No change required. The '`archive_name`' archival location is already configured for CloudOn.
            dict -- The full API response for `PATCH /internal/archive/object_store/{id}`.
        """

        valid_regions = ["westus", "westus2", "centralus", "eastus", "eastus2", "northcentralus", "southcentralus", "westcentralus", "canadacentral", "canadaeast", "brazilsouth", "northeurope", "westeurope",
                         "uksouth", "ukwest", "eastasia", "southeastasia", "japaneast", "japanwest", "australiaeast", "australiasoutheast", "centralindia", "southindia", "westindia", "koreacentral", "koreasouth"]

        if region not in valid_regions:
            sys.exit('Error: The `region` must be one of the following: {}'.format(valid_regions))

        self.log("azure_cloudon: Searching the Rubrik cluster for archival locations.")
        archives_on_cluster = self.get('internal', '/archive/object_store')

        config = {}
        config['name'] = archive_name
        config['objectStoreType'] = 'Azure'
        config['isComputeEnabled'] = True

        config['azureComputeSummary'] = {}
        config['azureComputeSummary']["tenantId"] = tenant_id
        config['azureComputeSummary']["subscriptionId"] = virtual_network_id.split("/")[2]
        config['azureComputeSummary']["clientId"] = application_id
        config['azureComputeSummary']["region"] = region
        config['azureComputeSummary']["generalPurposeStorageAccountName"] = storage_account_name
        config['azureComputeSummary']["containerName"] = container

        config['azureComputeSecret'] = {}
        config['azureComputeSecret']["clientSecret"] = application_key
        config['defaultComputeNetworkConfig'] = {}
        config['defaultComputeNetworkConfig']['subnetId'] = subnet_name
        config['defaultComputeNetworkConfig']['vNetId'] = virtual_network_id
        config['defaultComputeNetworkConfig']['securityGroupId'] = security_group_id

        redacted_archive_definition = {}
        redacted_archive_definition['name'] = archive_name
        redacted_archive_definition['objectStoreType'] = "Azure"
        redacted_archive_definition['accessKey'] = storage_account_name
        redacted_archive_definition['bucket'] = container
        redacted_archive_definition['isComputeEnabled'] = True

        redacted_archive_definition['azureComputeSummary'] = {}
        redacted_archive_definition['azureComputeSummary']["tenantId"] = tenant_id
        redacted_archive_definition['azureComputeSummary']["subscriptionId"] = virtual_network_id.split("/")[2]
        redacted_archive_definition['azureComputeSummary']["clientId"] = application_id
        redacted_archive_definition['azureComputeSummary']["region"] = region
        redacted_archive_definition['azureComputeSummary']["generalPurposeStorageAccountName"] = storage_account_name
        redacted_archive_definition['azureComputeSummary']["containerName"] = container

        redacted_archive_definition['defaultComputeNetworkConfig'] = {}
        redacted_archive_definition['defaultComputeNetworkConfig']['subnetId'] = subnet_name
        redacted_archive_definition['defaultComputeNetworkConfig']['vNetId'] = virtual_network_id
        redacted_archive_definition['defaultComputeNetworkConfig']['securityGroupId'] = security_group_id

        for archive in archives_on_cluster['data']:
            if archive['definition']['objectStoreType'] == 'Azure' and archive['definition']['name'] == archive_name:
                if archive['definition'] == redacted_archive_definition:
                    return "No change required. The '{}' archival location is already configured for CloudOn.".format(archive_name)
                else:
                    self.log("azure_cloudon: Updating the archive location for CloudOn.")
                    return self.patch('internal', "/archive/object_store/{}".format(archive['id']), config, timeout)

        sys.exit("Error: The Rubrik cluster does not have an archive location named '{}'.".format(archive_name))
